# This file will be sourced by the shell bash.
#
# Filename: integrity-database.bash
#
# Copyright (C) 2016-2021 Hartmut Buhrmester
#                         <wsusoffline-scripts-xxyh@hartmut-buhrmester.de>
#
# License
#
#     This file is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published
#     by the Free Software Foundation, either version 3 of the License,
#     or (at your option) any later version.
#
#     This file is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     General Public License for more details.
#
#     You should have received a copy of the GNU General
#     Public License along with this program.  If not, see
#     <http://www.gnu.org/licenses/>.
#
# Description
#
#     The integrity database is a set of hashdeep files in the directory
#     wsusoffline/client/md. Each hashes file corresponds to one download
#     directory, and each line in the hashes file is a fingerprint of
#     one downloaded file. In WSUS Offline Update, each fingerprint
#     consists of the file size, the MD5, SHA-1 and SHA-256 hashes and
#     the relative path to the md directory, separated with commas.
#
#     The integrity database is verified at the beginning of each
#     download run, then new files are downloaded and obsolete file are
#     removed. Finally, the hashdeep files will be recreated, to match
#     the new directory contents.
#
#     After creating the integrity database, the hashdeep files
#     can be used to verify the integrity of the downloaded
#     files: The SHA-1 hash is embedded into the filename of
#     all security updates, which are extracted from the WSUS
#     offline scan file wsusscn2.cab. For example, if the filename is
#     "ndp35sp1-kb958484-x64_e69006433c1006c53da651914dc8162bbdd80d41.exe",
#     then "e69006433c1006c53da651914dc8162bbdd80d41" is the SHA-1 hash,
#     consisting of 40 hexadecimal digits. hashdeep calculates the SHA-1
#     hashes itself, and then the calculated hashes can be compared to
#     those embedded into the filenames. If the hashes don't match,
#     the damaged files will be deleted, and the corresponding lines
#     excised from the hashdeep file.
#
#     This test is independent of verifying digital file signatures with
#     Sigcheck. It only uses hashdeep and works well on Linux.
#
#     The main purpose of the integrity database is probably the
#     verification of the updates, right before installing them. WSUS
#     Offline Update was once designed to create custom update
#     CDs/DVDs. Then only the installation part of WSUS Offline Update
#     is used - only the files and directories in the client directory
#     are copied to the ISO image file. This way, the client directory
#     of WSUS Offline Update becomes the root directory of the ISO
#     image. The image file is then burned to CDs/DVDs, and these media
#     are used to update the client machines. But recordable CDs/DVDs
#     are not very reliable. The integrity database can then be used to
#     verify the updates again, right before they are installed.
#
#     - https://forums.wsusoffline.net/viewtopic.php?f=3&t=6086
#
#
# TODO
#
#     The hashdeep files are created and verified from the client/md
#     directory. For the Linux download scripts, this requires constant
#     changes of the current working directory from sh to ../client/md
#     and back. The pathnames of the download directories must also
#     be shortened, before they are passed to hashdeep: the element
#     "client" must be removed, because it doesn't exist anymore during
#     installation.
#
#     This is similar in the Windows script DownloadUpdates.cmd: It
#     needs the same directory changes with pushd/popd.
#
#     These directory changes could be omitted by using the bare option
#     (-b) in hashdeep, which simply strips any leading directory
#     information. This should not cause problems, since in most
#     cases every hashdeep file corresponds to one download directory
#     only. Only for wddefs and msse (in the ESR version), there are
#     two subdirectories with the same filenames. Still, this should
#     not create problems, because hashdeep identifies files by the
#     hash first, not the filename. Of course, we could also create two
#     separate hashdeep files for the two subdirectories, as it was done
#     for the dotnet subdirectories.

# ========== Configuration ================================================

# In the Windows script DownloadUpdates.cmd, hashdeep always uses
# three hash functions: MD5, SHA-1 and SHA-256. This looks a bit
# over-engineered, and it makes the creation and validation of the
# integrity database rather slow on old machines. The Linux download
# scripts 1.19.4-ESR and 2.3 introduced a new optional "fast mode",
# which only calculates the SHA-1 hash.
#
# As with other permanent settings, the default value is provided here,
# but it should be changed in the file preferences.bash.

fast_mode="${fast_mode:-disabled}"

# ========== Functions ====================================================

# function verify_integrity_database
#
# The existing files in the ../client directory are usually verified
# before downloading new files.
#
# Parameter 1: The hashed directory with a relative path,
#              e.g. ../client/ofc/glb
# Parameter 2: The hashdeep file with a relative path,
#              e.g. ../client/md/hashes-ofc-glb.txt
#
# Before calling hashdeep, the function changes to the directory
# ../client/md. It immediately switches back to the previous directory
# after calling hashdeep. This implies for hashdeep:
#
# - the relative path of the hashed directory must be shortened,
#   e.g. to ../ofc/glb
# - the hashdeep file can be referenced with just the filename,
#   e.g. hashes-ofc-glb.txt
# - the logfile must be referenced with an absolute path
#
# This approach is also used for the function create_integrity_database.

function verify_integrity_database ()
{
    local hashed_dir="$1"
    local hashed_dir_shortened="${hashed_dir/'client/'/}"
    local hashes_pathname="$2"
    local hashes_filename="${hashes_pathname##*/}"
    local -a file_list=()
    local -a hashdeep_input=()

    mkdir -p "../client/md"

    # Preconditions
    #
    # If the integrity database is disabled in the file preferences.bash,
    # then existing hashdeep files should be removed, because they will
    # become invalid after downloading new and deleting old files.
    # shellcheck disable=SC2154
    if [[ "${use_integrity_database}" == "disabled" ]]
    then
        log_info_message "Verification of the integrity database is disabled in preferences.bash"
        rm -f "${hashes_pathname}"
        return 0
    fi

    log_info_message "Verifying the integrity of existing files in the directory ${hashed_dir} ..."

    # When the script is first run, neither the hashed directory nor
    # the hashes file exist yet. This is not an error; it only means,
    # that the download task must be run once to create the file.
    if ! require_directory "${hashed_dir}"
    then
        log_info_message "The download directory ${hashed_dir} does not exist yet. This is normal during the first run of the script."
        # If the download directory is missing, then the hashdeep file
        # should not exist either.
        if [[ -f "${hashes_pathname}" ]]
        then
            log_warning_message "The file ${hashes_filename} is not valid and will be deleted..."
            rm "${hashes_pathname}"
        fi
        return 0
    fi
    # The function ensure_non_empty_file will delete spurious empty files
    if ! ensure_non_empty_file "${hashes_pathname}"
    then
        log_info_message "The hashdeep file ${hashes_filename} does not exist yet. This is normal during the first run of the script."
        return 0
    fi

    # Count the input files in the hashed directory. If no input files
    # are found, then the verification should be aborted at this point.
    #
    # This check is important for the directory ../client/dotnet: It
    # is not scanned recursively, because the subdirectories x86-glb
    # and x64-glb should not be included in the hashes file. A list of
    # filenames is created instead and passed to hashdeep. If this list
    # is empty, then hashdeep won't find any input files and read from
    # standard input instead. This means, that the script may hang at
    # this point.
    #
    # This does not happen, if an empty directory is passed to hashdeep,
    # but then all files listed in the hashdeep file would be listed
    # as missing.
    case "${hashed_dir}" in
        ../client/dotnet )
            shopt -s nullglob
            # Static installers are often .exe files, but the file list
            # should not necessarily be restricted to one file type
            file_list=( ../client/dotnet/*.* )
            shopt -u nullglob
        ;;
        ../client/msse | ../client/wddefs )
            shopt -s nullglob
            file_list=(
                "${hashed_dir}/x86-glb"/*.*
                "${hashed_dir}/x64-glb"/*.*
            )
            shopt -u nullglob
        ;;
        *)
            shopt -s nullglob
            file_list=( "${hashed_dir}"/*.* )
            shopt -u nullglob
        ;;
    esac
    if (( "${#file_list[@]}" == 0 ))
    then
        log_file_integrity_verification_error "The verification of existing files was aborted, because there are no input files in the directory ${hashed_dir}."
        increment_error_count
        if [[ -f "${hashes_pathname}" ]]
        then
            log_warning_message "The file ${hashes_filename} is not valid and will be deleted..."
            rm "${hashes_pathname}"
        fi
        return 0
    fi

    # Create a copy of the hashes file in Linux format
    # shellcheck disable=SC1003,2154
    cat_dos "${hashes_pathname}" | tr '\\' '/' > "${temp_dir}/${hashes_filename}"

    # The hashdeep input can be a list of files or a single directory
    if [[ "${hashed_dir}" == "../client/dotnet" ]]
    then
        pushd "../client/md" &> /dev/null || fail "pushd failed"
        shopt -s nullglob
        # Static installers are often .exe files, but the file list
        # should not necessarily be restricted to one file type
        hashdeep_input=( ../dotnet/*.* )
        shopt -u nullglob
        popd &> /dev/null || fail "popd failed"
    else
        hashdeep_input=( -r "${hashed_dir_shortened}" )
    fi

    # Redirecting the hashdeep output to the logfile after changing the
    # current working directory requires, that the logfile is defined
    # with an absolute path. A relative path like "../log/download.log"
    # won't work anymore.
    #
    # This was changed in the Linux download scripts version 1.11,
    # with hashdeep and curl in mind.
    pushd "../client/md" &> /dev/null || fail "pushd failed"
    # shellcheck disable=SC2154
    if hashdeep -a -vv -k "${temp_dir}/${hashes_filename}"  \
                -l "${hashdeep_input[@]}" 2>&1              \
        | tee -a "${logfile}"
    then
        log_info_message "Verified the integrity of existing files in the directory ${hashed_dir}"
    else
        log_file_integrity_verification_error "The directory ${hashed_dir} has changed since last creating the integrity database"
        increment_error_count
        log_warning_message "The file ${hashes_filename} is not valid and will be deleted..."
        rm "${hashes_filename}"
    fi
    popd &> /dev/null || fail "popd failed"

    # Integrity database verification errors are only reported and
    # written to the logfile. The script may just go on, because the
    # hashdeep files will be deleted and rebuilt anyway.
    return 0
}


function create_integrity_database ()
{
    local hashed_dir="$1"
    local hashed_dir_shortened="${hashed_dir/'client/'/}"
    local hashes_pathname="$2"
    local hashes_filename="${hashes_pathname##*/}"
    local -a file_list=()
    local -a hashdeep_input=()
    # The list of hash functions, that hashdeep uses
    local hash_functions=""
    if [[ "${fast_mode}" == "enabled" ]]
    then
        hash_functions="sha1"
    else
        hash_functions="md5,sha1,sha256"
    fi

    mkdir -p "../client/md"

    # Preconditions
    if [[ "${use_integrity_database}" == "disabled" ]]
    then
        log_info_message "Creation of the integrity database is disabled in preferences.bash"
        return 0
    fi

    log_info_message "Creating integrity database for directory ${hashed_dir} ..."
    # Delete existing hashdeep files
    rm -f "${hashes_pathname}"

    if ! require_directory "${hashed_dir}"
    then
        log_warning_message "The creation of the integrity database was aborted, because the directory ${hashed_dir} does not exist."
        return 0
    fi

    # Count files in the hashed directory to prevent the creation of
    # empty hashdeep files and other problems.
    #
    # For the directory ../client/dotnet, a list of filenames is created
    # and passed to hashdeep. If the list is empty, then hashdeep won't
    # find any input files and read from standard input instead. This
    # means, that the script may hang at this point.
    #
    # This does not happen, if an empty directory is passed to hashdeep,
    # but if hashdeep doesn't find any input files, then it won't produce
    # any output. The output redirection will still create an empty file.
    #
    # For some time, this was one of the most common errors with hashdeep
    # in the Windows script DownloadUpdates.cmd.
    #
    # - https://forums.wsusoffline.net/viewtopic.php?f=5&t=5336
    case "${hashed_dir}" in
        ../client/dotnet )
            shopt -s nullglob
            file_list=( ../client/dotnet/*.* )
            shopt -u nullglob
        ;;
        ../client/msse | ../client/wddefs )
            shopt -s nullglob
            file_list=(
                "${hashed_dir}/x86-glb"/*.*
                "${hashed_dir}/x64-glb"/*.*
            )
            shopt -u nullglob
        ;;
        *)
            shopt -s nullglob
            file_list=( "${hashed_dir}"/*.* )
            shopt -u nullglob
        ;;
    esac
    if (( "${#file_list[@]}" == 0 ))
    then
        log_warning_message "The creation of the integrity database was aborted, because there are no input files in the directory ${hashed_dir}."
        return 0
    fi

    # The hashdeep input can be a list of files or a single directory
    if [[ "${hashed_dir}" == "../client/dotnet" ]]
    then
        pushd "../client/md" &> /dev/null || fail "pushd failed"
        shopt -s nullglob
        hashdeep_input=( ../dotnet/*.* )
        shopt -u nullglob
        popd &> /dev/null || fail "popd failed"
    else
        hashdeep_input=( -r "${hashed_dir_shortened}" )
    fi

    # The hashdeep standard output is written to the hashes file,
    # but error output can be duplicated with tee and written to the
    # terminal and the logfile.
    pushd "../client/md" &> /dev/null || fail "pushd failed"
    # shellcheck disable=SC1003
    if { hashdeep -c "${hash_functions}" -l "${hashdeep_input[@]}"  \
            | tr '/' '\\'                                           \
            | unix_to_dos                                           \
            > "${hashes_filename}"
       } 2>&1                                                       \
            | tee -a "${logfile}"
    then
        # Output redirections may create empty files, even if hashdeep
        # does not indicate any error
        if ensure_non_empty_file "${hashes_filename}"
        then
            log_info_message "Created file ${hashes_filename}"
        fi
    else
        log_error_message "Creation of hashdeep file ${hashes_filename} failed"
        increment_error_count
        rm -f "${hashes_filename}"
    fi
    popd &> /dev/null || fail "popd failed"

    return 0
}


# function verify_embedded_checksums
#
# This function checks the integrity of the downloaded files by comparing
# the SHA-1 hashes, which are embedded into the filenames of most security
# updates, with the calculated values.
#
# Hashdeep files are CSV-formatted text files. The default field order
# in WSUS Offline Update is:
#
# field 1 = file size
# field 2 = MD5 hash
# field 3 = SHA-1 hash
# field 4 = SHA-256 hash
# field 5 = relative file path
#
# With the optional "fast mode", the field order is:
#
# field 1 = file size
# field 2 = SHA-1 hash
# field 3 = relative file path
#
# When reading the hashdeep file, the first five lines with comments are
# skipped. The remaining lines are split into five fields (the last two
# fields may be empty).
#
# Then the variables sha1_calculated and windows_path are assigned using
# name references:
#
# - sha1_calculated will point to field_3 in default mode, but to field_2
#   in fast mode
# - windows_path will point to field_5 in the default mode, but to
#   field_3 in fast mode

function verify_embedded_checksums ()
{
    local hashed_dir="$1"
    local hashed_dir_shortened="${hashed_dir/'client/'/}"
    local hashes_pathname="$2"
    local hashes_filename="${hashes_pathname##*/}"

    local field_1=""
    local field_2=""
    local field_3=""
    local field_4=""
    local field_5=""
    # TODO: sha1_calculated and windows_path should probably be declared
    # with the option -n, to become automatic name references, but
    # "${!sha1_calculated}" works as well.
    local sha1_calculated=""  # SHA-1 hash calculated by hashdeep
    local windows_path=""     # relative file path in the hashes file
    if [[ "${fast_mode}" == "enabled" ]]
    then
        sha1_calculated="field_2"
        windows_path="field_3"
    else
        sha1_calculated="field_3"
        windows_path="field_5"
    fi
    local sha1_embedded=""  # SHA-1 hash embedded into the filename
    local linux_path        # the relative path with forward slashes "/"
    local -i initial_errors="0"
    initial_errors="$(get_error_count)"

    mkdir -p "../client/md"

    # Preconditions
    if [[ "${use_integrity_database}" == "disabled" ]]
    then
        log_info_message "Verification of embedded hashes is disabled in preferences.bash"
        return 0
    fi

    log_info_message "Verifying embedded SHA1 hashes for directory ${hashed_dir} ..."

    if ! require_directory "${hashed_dir}"
    then
        log_warning_message "The verification of embedded hashes was aborted, because the directory ${hashed_dir} does not exist."
        if [[ -f "${hashes_pathname}" ]]
        then
            log_warning_message "The file ${hashes_filename} is not valid and will be deleted..."
            rm "${hashes_pathname}"
        fi
        return 0
    fi
    if ! require_non_empty_file "${hashes_pathname}"
    then
        log_warning_message "The hashdeep file ${hashes_filename} does not exist."
        return 0
    fi

    # Extract filenames with embedded SHA-1 hashes from the hashdeep
    # file. The search pattern consists of an underscore, the SHA-1 hash
    # (40 hexadecimal digits) and the file extension.
    #
    # The first five lines of the hashdeep file are skipped, because
    # they only contain comments.
    #
    # Carriage returns are removed, but the path delimiters are not
    # changed at this point. The original Windows path may be needed
    # later, to delete invalid records from the hashdeep file.
    tail_dos -n +6 "${hashes_pathname}"                 \
        | grep -E '_[[:xdigit:]]{40}[.][[:alpha:]]{3}'  \
        > "${temp_dir}/sha-1-${hashes_filename}"        \
        || true

    # Tell shellcheck to not complain about "unused variables"
    # shellcheck disable=SC2034
    while IFS=',' read -r field_1 field_2 field_3 field_4 field_5
    do
        # Extract the embedded SHA-1 hash from the pathname. GNU grep
        # could conveniently use:
        #
        # grep -E --only-matching '[[:xdigit:]]{40}'
        #
        # but this was replaced with sed for compatibility. Extended
        # regular expressions work the same as basic regular expressions,
        # but parentheses and curly brackets don't need to be escaped.
        #sha1_embedded="$( sed 's/.*_\([[:xdigit:]]\{40\}\).*/\1/g' <<< "${!windows_path}" || true )"
        sha1_embedded="$( sed -E 's/.*_([[:xdigit:]]{40}).*/\1/' <<< "${!windows_path}" || true )"
        # Compare the embedded SHA-1 hashes with those calculated
        # by hashdeep
        if [[ "${!sha1_calculated}" != "${sha1_embedded}" ]]
        then
            # Change path delimiters from Windows to Linux
            # shellcheck disable=SC1003
            linux_path="${!windows_path//'\'//}"
            log_error_message "Trashing/deleting file ${linux_path##*/} due to mismatching SHA-1 message digests..."
            increment_error_count
            # The file paths in the hashdeep files are relative to the
            # directory ../client/md
            pushd "../client/md" &> /dev/null || fail "pushd failed"
            trash_file "${linux_path}"
            popd &> /dev/null || fail "cd failed"

            # Rewrite the original hashes file (not the copy in the
            # temporary directory) without the deleted file.
            #
            # This step searches for the original Windows path in the
            # hashdeep file, because the SHA-1 hash may be ambiguous: If
            # static and dynamic download links from different sources
            # are mixed, then the same file may be downloaded twice,
            # but with different file names.
            #
            # The Windows path can only be searched literally, e.g. with
            # grep -F. In regular expressions, all backslashes need to
            # be escaped. This also prevents the use of escape sequences
            # like \r for carriage returns.
            #
            # TODO: This step may also delete backup
            # files from the hashdeep file, if the
            # beginning of the filename is the same, e.g. both
            # ndp45-kb4014566-x86_03cfe905211220f70c6d9eeb61dfdbfa12897519.exe
            # and
            # ndp45-kb4014566-x86_03cfe905211220f70c6d9eeb61dfdbfa12897519.exe.bak
            # might be deleted. However, this is a constructed example,
            # which doesn't appear in normal use, and such problems are
            # corrected on the next download run.
            #
            # There may be other solutions to try: sed can delete lines
            # directly with the command "d".
            mv "${hashes_pathname}" "${hashes_pathname}.bak"
            grep -F -v "${!windows_path}" "${hashes_pathname}.bak"  \
                > "${hashes_pathname}"                              \
                || true
            rm "${hashes_pathname}.bak"
        fi
    done < "${temp_dir}/sha-1-${hashes_filename}"

    if same_error_count "${initial_errors}"
    then
        log_info_message "Verified embedded SHA1 hashes"
    else
        log_warning_message "Verification of embedded SHA1 hashes detected $(get_error_difference "${initial_errors}") errors"
    fi
    return 0
}

return 0
